#include <string.h>
#include "LogParser.h"
#include "DynLista.h"

/* #define DEBUG */

void ListInit(struct ListHead* head, char IDname)
{
  if(head==NULL)
    {
      fprintf(stderr, "ListInit: head==NULL, bye!\n");
      exit(1);
    }

  head->namecount=0;
  head->ID=IDname;
  head->First=NULL;
  head->Last=NULL;
}

unsigned long NameInList(char *name, struct ListHead* head)
{ /* returns 0 if not found, position (starting at 1) if found */
  struct Node* node;
  int found;
  long pos;
  
  if(head==NULL)
    {
      fprintf(stderr, "NameInList: head==NULL, bye!\n");
      exit(1);
    }

  pos=0;
  found=FALSE;
  node=head->First;

  while((node!=NULL) && (found!=TRUE))
    {
      pos++;
      if(node->name==NULL)
	fprintf(stderr, "NIL ERR: node->name==NULL\n");
      if(name==NULL)
	fprintf(stderr, "NIL ERR: name==NULL\n");
      
      if(strncasecmp(node->name, name, strlen(name))==0)
	{
	  found=TRUE;
	}
      else
	{
	  node=node->next;
	}
    }
  if(found==FALSE)
    {
      pos=0;
    }

  return pos;
}


void AddNameToList(char *src, struct ListHead* head)
{
  struct Node* t;
  void* blah;
  size_t nodesize;

  /* Verify parameters */
  if(src==NULL)
    {
      fprintf(stderr, "AddNameToList: src==NULL, bye!\n");
      exit(1);
    }
  if(head==NULL)
    {
      fprintf(stderr, "AddNameToList: head==NULL, bye!\n");
      exit(1);
    }

  /* get memory and build a node */
  nodesize=sizeof(struct Node);
  blah=malloc(nodesize);
  t=(struct Node*)blah;

  if(t==NULL)
    {
      fprintf(stderr, "AddNameToList failed to allocate memory (1).\n");
      exit(1);
    }
  t->hitcount=1;
  t->next=NULL;

  blah=malloc(strlen(src)+1);
  t->name=(char*)blah;

  if(blah==NULL)
    {
      fprintf(stderr, "AddNameToList failed to allocate memory (2).\n");
      exit(1);
    }
  strcpy(t->name, src);
  
  /* add node to list */
  if(head->First==NULL)
    { /* list is empty */
      head->First=t;
      t->prev=NULL;
    }
  else
    { /* list contains at least one element */
      t->prev=head->Last;
      head->Last->next=t;
    }
  head->Last=t;
  head->namecount++;
}

void EmptyList(struct ListHead* head)
{
  struct Node *curr, *next;
  
  if(head==NULL)
    {
      fprintf(stderr, "EmptyList: head==NULL, bye!\n");
      exit(1);
    }

  curr=head->First;
  while(curr!=NULL)
    {
      next=curr->next;
      free(curr);
      curr=next;
    }
  head->First=head->Last=NULL;
  head->namecount=0;
}

struct Node* GetElement(struct ListHead* head, unsigned long pos)
{ /* returns null if pos is beyond the list */
  struct Node* node;
  int found;
  unsigned long i;

  if(head==NULL)
    {
      fprintf(stderr, "GetElement: head==NULL, bye!\n");
      exit(1);
    }
  if(pos==0)
    {
      fprintf(stderr, "GetElement: pos==0, bye!\n");
      exit(1);
    }

  i=0;
  found=FALSE;
  node=head->First;
  while((node!=NULL) && (found!=TRUE))
    {
      i++;
      if(pos==i)
	{
	  found=TRUE;
	}
      else
	{
	  node=node->next;
	}
    }
  if(found==FALSE)
    return NULL;
  else
    return node;
}

void SwapNodes(struct ListHead* head, struct Node *n1, struct Node *n2)
{ 
  struct Node t;

  if(head==NULL)
    {
      fprintf(stderr, "SwapNodes: head==NULL, bye!\n");
      exit(1);
    }
  if(n1==NULL)
    {
      fprintf(stderr, "SwapNodes: n1==NULL, bye!\n");
      exit(1);
    }
  if(n2==NULL)
    {
      fprintf(stderr, "SwapNodes: n2==NULL, bye!\n");
      exit(1);
    }

  /* should probably only swap some pointers but I don't care right now... */

  /* Copy data from n1 to t */
  t.name=n1->name;
  t.hitcount=n1->hitcount;
  t.printed=n1->printed;

  /* Copy data from n2 to n1 */
  n1->name=n2->name;
  n1->hitcount=n2->hitcount;
  n1->printed=n2->printed;

  /* Copy data from t to n2 */
  n2->name=t.name;
  n2->hitcount=t.hitcount;
  n2->printed=t.printed;
}

void SortNames(struct ListHead* head)
{ /* sort the list somehow, insertionsort will do for now */
  struct Node *current, *searcher, *high;
  unsigned long hiscore;

  if(head==NULL)
    {
      fprintf(stderr, "SortNames: head==NULL, bye!\n");
      exit(1);
    }

  current=head->First;
  if(current==NULL)
    {
      fprintf(stderr, "SortNames: current==NULL, bye!\n");
      exit(1);
    }

  while(current->next!=NULL)
    {
      high=current; /* does not matter but removes a warning */
      hiscore=current->hitcount;
      searcher=current->next;
      while(searcher!=NULL)
	{
	  if(searcher->hitcount > hiscore)
	    {
	      high=searcher;
	      hiscore=high->hitcount;
	    }
	  searcher=searcher->next;
	}
      if(hiscore > current->hitcount)
	SwapNodes(head, current, high);
      current=current->next;
    }
}

void IncCount(struct ListHead* head, unsigned long place)
{
  struct Node *node;

  if(head==NULL)
    {
      fprintf(stderr, "IncCount: head==NULL, bye!\n");
      exit(1);
    }
  if(place==0)
    {
      fprintf(stderr, "IncCount: place==0, bye!\n");
      exit(1);
    }

  node=GetElement(head, place);
  if(node==0)
    {
      fprintf(stderr, "IncCount: node==0, bye!\n");
      exit(1);
    }
  node->hitcount++;
}


int dyntest_main(void)
{
  struct ListHead namnlista;
  struct Node *test;
  unsigned long pos;

  ListInit(&namnlista, 'D'); /* D as in Dummy */
  
  AddNameToList("namnet", &namnlista);
  AddNameToList("namnet2", &namnlista);
  AddNameToList("namnet3", &namnlista);
  AddNameToList("namnet4", &namnlista);
  AddNameToList("namnet5", &namnlista);
  AddNameToList("namnet6", &namnlista);
  AddNameToList("namnet7", &namnlista);
  AddNameToList("namnet8", &namnlista);
  AddNameToList("namnet9", &namnlista);
  AddNameToList("namnet10", &namnlista);
  AddNameToList("namnet11", &namnlista);
  AddNameToList("namnet12", &namnlista);
  AddNameToList("namnet13", &namnlista);
  AddNameToList("namnet14", &namnlista);
  AddNameToList("namnet15", &namnlista);
  AddNameToList("namnet16", &namnlista);
  AddNameToList("namnet17", &namnlista);
  AddNameToList("namnet18", &namnlista);
  AddNameToList("namnet19", &namnlista);
  AddNameToList("namnet20", &namnlista);

  IncCount(&namnlista, 1);
  IncCount(&namnlista, 1);
  IncCount(&namnlista, 1);
  IncCount(&namnlista, 1);
  IncCount(&namnlista, 1);
  IncCount(&namnlista, 1);
  IncCount(&namnlista, 1);
  IncCount(&namnlista, 10);
  IncCount(&namnlista, 2);
  IncCount(&namnlista, 10);
  IncCount(&namnlista, 10);

  printf("Dumpar:\n");
/*   test=namnlista.First; */
/*   pos=1; */
/*   while(test!=NULL) */
/*     { */
/*       printf("%d: %s (%ld)\n", pos, test->name, test->hitcount); */
/*       pos++; */
/*       test=test->next; */
/*     } */

  for(pos=1;pos<=namnlista.namecount;pos++)
    {
      test=GetElement(&namnlista, pos);
      printf("%ld: %s (%ld)\n", pos, test->name, test->hitcount);
    }
  EmptyList(&namnlista);
  
  return 0;
}

unsigned long FindMaxCountList(struct ListHead* head)
{
  struct Node* node;
  unsigned long i;

  if(head==NULL)
    {
      fprintf(stderr, "FindMaxCountList: head==NULL, bye!\n");
      exit(1);
    }

  i=0;
  node=head->First;
  while(node!=NULL)
    {
      if(node->hitcount>i)
	i=node->hitcount;
      node=node->next;
    }
  return i;
}
