/* ********************************************************** */
/* ********     Copyright by Jonas Svensson      ************ */
/* ********            960227 - 971028           ************ */
/* ********         jonass@lysator.liu.se        ************ */
/* ******** <http://www.lysator.liu.se/~jonass/> ************ */
/* ********    a.k.a.<http://travel.to/space>    ************ */
/* ********************************************************** */
/* ****  This program is freeware. That is you may copy ***** */
/* ****  and use it free of charge.                     ***** */
/* ****  You are not allowed to modify it or use parts  ***** */
/* ****  of it in your own programs without permission. ***** */
/* ****  Nor may you sell or distribute it at any cost  ***** */
/* ****  without permission.                            ***** */
/* ****  Neither is the author responsible for any      ***** */
/* ****  kind of damage caused by this program, you     ***** */
/* ****  use it entirely without guarantee althou the   ***** */
/* ****  the author believes it is correct and will     ***** */
/* ****  not cause any damage.                          ***** */
/* ********************************************************** */


/* This program will create a "html" file with some statistics
   calculated from the logfile created by a httpd (like spinner).
   
   Bugreports and other comments to: jonass@lysator.liu.se

   Comments like "I compiled it succesfully on this or that system,
   with this commandline: xxxxxxxxxxx" are most welcome.

   HOWTO:
     Find the file LogParser.ini and change the data (if necessary).
     Then place it in your home directory and call it .logparserrc
     and change the define of RCNAME to suit your needs.

     Just type make to compile. If your system is not supported by
     the included makefile, please modify it and send me a copy.

     "cc" might work instead of "gcc" on some systems.

   HISTORY:
   0.99 ------ Collection of all private versions.
   1.00 960327
   1.01 960327 (also!) makes links of referer and domains.
   1.02 960330 bug: wrong format on the number of hits
   1.03 960421 Uses PRE instead of TT for the "graph"
   1.04 960425 Filestat and such
   1.05 960506 percentage instead of absolut numbers
   1.06 960516 created a makefile and use of inimodule
               now prints to a specified file (ie not stdout)
   1.07 960524 and you know, the file should be closed afterwards.
	       possible to set output filename to stdout
	       userdefined number of entries on the top lists
	       Separated Browser and OS (or whatever)
	       Domains & TopDomains
   1.08 96xxxx tries more to keep out of unallocated memory
               number 0 will not show any entries of a kind
	       top list for referersites added
   1.09 960604 Cleaned up 1.08 by removing some bugs...
   1.10 960630 DynLista and some minor things
   1.20 971023 New logfile-format some other things (never made it thru the door)
   1.21 971028 Bugs!Bugs!Bugs! Will they ever leave?
   1.22 971029 Accesses by date
   1.23 971105 Dynamic allocation of memory for node-data in DynLista.
               Should use less memory and sometimes be faster.

   TODO:
   - Pretty graphs is quite low on priority.
   - bugs ? done.
   - access by day of month
   - access by week
   - referer kan innehlla space (sk: "file:" eller "FIN")
   - fixa bttre stt att hitta strngarna t.ex. en state-machine som splittar allt i ett
*/

#include "inimodule.h"

#define RCNAME ".logparserrc"

/* #define DEBUG */

#define PRG_NAME "W3-LogParser v1.23"

#include "LogParser.h"
#include "DynLista.h"
#include <string.h>

char LoggFileName[MAX_FILENAME_LENGTH];
char PageName[MAX_LINE_LENGTH];
char CreateFileName[MAX_LINE_LENGTH];
char buff[MAX_LINE_LENGTH];

char *numerical="numerical";
char *noOS="No OS";

char HostNameTemp[MAX_LENGTH_OF_HOSTNAME];
char temp[MAX_LINE_LENGTH]; /* life span about 3-5 lines */

unsigned long StartYear, StartMonth, StartDay, StartHour, StartMinute;
unsigned long EndYear, EndMonth, EndDay, EndHour, EndMinute;
unsigned long LinesInLog;

unsigned long TimedAccess[24], MonthAccess[12];

struct ListHead host, referer, browser, domain, command, file, OSname,
       TopDomains, Referersites, mdays;
int ShowMaxHosts, ShowMaxReferers, ShowMaxBrowsers, ShowMaxDomains, 
    ShowMaxCommands, ShowMaxFiles, ShowMaxOSnames, ShowMaxTopDomains,
    ShowMaxReferersites;

void init(void)
{
  int i;
  char filnamn[MAX_FILENAME_LENGTH], *homepek;
  FILE *fp;
  
  fp=fopen(RCNAME, "r");
  if(fp!=NULL)
    {
      fclose(fp);
      strcpy(filnamn, RCNAME);
    }
  else
    {  
      homepek=getenv("HOME");
      strcpy(filnamn, homepek);
      strcat(filnamn, "/");
      strcat(filnamn, RCNAME);
    }
  
fprintf(stderr, "inifile=%s\n", filnamn);

  i=GetPrivateProfileString("USER", "LogFileName", "AccessLog", LoggFileName, sizeof(LoggFileName), filnamn);
  i=GetPrivateProfileString("USER", "PageName", "www.lysator.liu.se/~jonass/stat.html", PageName, sizeof(PageName), filnamn);
  i=GetPrivateProfileString("USER", "OutputFileName", "stat.html", CreateFileName, sizeof(CreateFileName), filnamn);
  
  ShowMaxHosts=GetPrivateProfileInt("USER", "Hosts", 20, filnamn);
  ShowMaxReferers=GetPrivateProfileInt("USER", "Referers", 20, filnamn);
  ShowMaxBrowsers=GetPrivateProfileInt("USER", "Browsers", 20, filnamn);
  ShowMaxDomains=GetPrivateProfileInt("USER", "Domains", 20, filnamn);
  ShowMaxCommands=GetPrivateProfileInt("USER", "Commands", 20, filnamn);
  ShowMaxFiles=GetPrivateProfileInt("USER", "Files", 20, filnamn);
  ShowMaxOSnames=GetPrivateProfileInt("USER", "OSnames", 20, filnamn);
  ShowMaxTopDomains=GetPrivateProfileInt("USER", "TopDomains", 20, filnamn);
  ShowMaxReferersites=GetPrivateProfileInt("USER", "Referersites", 20, filnamn);
  
  
  StartYear=40000; /*Should be enough for the life expectancy of this program...*/
  EndYear=0;
  
  ListInit(&host, 'h');
  ListInit(&referer, 'r');
  ListInit(&browser, 'b');
  ListInit(&domain, 'd');
  ListInit(&command, 'c');
  ListInit(&file, 'f');
  ListInit(&OSname, 'O');
  ListInit(&TopDomains, 'T');
  ListInit(&Referersites, 'R');
  ListInit(&mdays, 'y');
  
  i=0;
  while(i<24)
    {
      TimedAccess[i]=0;
      i++;
    }
  
  i=0;
  while(i<12)
    {
      MonthAccess[i]=0;
      i++;
    }
  
  LinesInLog=0;
  sethostent(1); /* to keep the connection to nameserver open */
}

void clean(void)
{
  endhostent();  /* close the connection to nameserver */
  EmptyList(&host);
  EmptyList(&referer);
  EmptyList(&browser);
  EmptyList(&domain);
  EmptyList(&command);
  EmptyList(&file);
  EmptyList(&OSname);
  EmptyList(&TopDomains);
  EmptyList(&Referersites);
}

/* Berkna JD fr en viss dag. Om man vill vara petig, s r det JD fr
   klockan 12:00 (mitt p dagen) som berknas, eftersom JD ligger precis
   i "motfas" med det vanliga kalendersystemet...
   Algoritm: Meeus, Jean, Astronomical Formulae for Calculators, 2 ed, s 24
*/

ulong date2jd(ulong y, ulong m, ulong d)
{
    ulong a;
    ulong jd;

    if (m<3)
      {
	  y--;
	  m+=12;
      }
    a = y/100;
    jd = 1720995 + d + 2 - a + (a/4) +
         (36525*y)/100 + (306001*(m+1))/10000;
    return jd;
}

char HexToChar(char *pek)
/*     *pek & *(pek+1) is supposed to be isxdigit()   */
{
  unsigned char t;
  
  t=0;
  switch (*pek)
    {
    case '0' : t=0; break;
    case '1' : t=1*16; break;
    case '2' : t=2*16; break;
    case '3' : t=3*16; break;
    case '4' : t=4*16; break;
    case '5' : t=5*16; break;
    case '6' : t=6*16; break;
    case '7' : t=7*16; break;
    case '8' : t=8*16; break;
    case '9' : t=9*16; break;
    case 'a' :
    case 'A' : t=10*16; break;
    case 'b' :
    case 'B' : t=11*16; break;
    case 'c' :
    case 'C' : t=12*16; break;
    case 'd' :
    case 'D' : t=13*16; break;
    case 'e' :
    case 'E' : t=14*16; break;
    case 'f' :
    case 'F' : t=15*16; break;
    }

  switch (*(pek+1))
    {
    case '1' : t+=1; break;
    case '2' : t+=2; break;
    case '3' : t+=3; break;
    case '4' : t+=4; break;
    case '5' : t+=5; break;
    case '6' : t+=6; break;
    case '7' : t+=7; break;
    case '8' : t+=8; break;
    case '9' : t+=9; break;
    case 'a' :
    case 'A' : t+=10; break;
    case 'b' :
    case 'B' : t+=11; break;
    case 'c' :
    case 'C' : t+=12; break;
    case 'd' :
    case 'D' : t+=13; break;
    case 'e' :
    case 'E' : t+=14; break;
    case 'f' :
    case 'F' : t+=15; break;
    }
  return t;
}

void htmlprint(FILE *fp, char *pek, int maxlength)
{
  int numme;

  numme=0;

  while((*pek!='\0') && (numme < maxlength))
    {
      if(*pek!='%')
	{
	  fprintf(fp, "%c", *pek);
	  pek++;
	  numme++;
	}
      else
	{  /* nt som brjar p % */
	  pek++;
	  numme++;
	  if(*pek=='%')
	    {                                                /*    %%   =>   %    */
	      fprintf(fp, "%%");
	      pek++;
	      numme++;
	    }
	  else
	    {
	      if(isxdigit(*pek) && (isxdigit(*(pek+1))))
		/* Two digits equals a special char like space, tilde, quotes or such. */
		{
		  fprintf(fp, "%c", HexToChar(pek));
		  pek+=2;
		  numme+=2;
		}
	      else
		{ /* Unknown sequence starting with % */
		  fprintf(fp, "%%");
		  /* print the % and do a fresh start... */
		}
	    }
	}
    }
  if(numme==maxlength)
    fprintf(fp, "...");
}

int HostExists(char *TestName)
{
  if(gethostbyname(TestName)==NULL)
    return FALSE;
  else
    return TRUE;
}

void MyMonthExpander(FILE *fp, int i)
{
  switch(i)
    {
    case 0 : fprintf(fp, "Jan   "); break;
    case 1 : fprintf(fp, "Feb   "); break;
    case 2 : fprintf(fp, "Mar   "); break;
    case 3 : fprintf(fp, "Apr   "); break;
    case 4 : fprintf(fp, "May   "); break;
    case 5 : fprintf(fp, "Jun   "); break;
    case 6 : fprintf(fp, "Jul   "); break;
    case 7 : fprintf(fp, "Aug   "); break;
    case 8 : fprintf(fp, "Sep   "); break;
    case 9 : fprintf(fp, "Oct   "); break;
    case 10 : fprintf(fp, "Nov   "); break;
    case 11 : fprintf(fp, "Dec   "); break;
    }
}

void MyPrintSummary(void)
{
  int i;
  time_t tid;
  ulong days;
  unsigned long max, value, j;
  char TestName[MAX_LENGTH_OF_HOSTNAME];
  FILE *fp;
  struct Node *node;
  struct Node* nodep;

#ifdef DEBUG
  fprintf(stderr, "MYP\n");
#endif
  if(strcasecmp("stdout", CreateFileName)==0)
    fp=stdout;
  else
    {
      fp=fopen(CreateFileName, "w");
      if(fp==NULL)
	{
	  fprintf(stderr, "Could not open %s for writing\n", CreateFileName);
	  return;
	}
    }
  
/* header */
  fprintf(fp, "<html> <head> <title>Created by %s</title></head>\n", PRG_NAME);
  fprintf(fp, "<body><h1>Log statistics</h1>\n");
/* end of header */

  /* could use difftime instead of date2jd ? */
  days=date2jd(EndYear, EndMonth, EndDay)-date2jd(StartYear, StartMonth, StartDay)+1;
  fprintf(fp, "Log spans from %4lu-%02lu-%02lu %2lu:%02lu to %4lu-%02lu-%02lu %2lu:%02lu (%lu days).<br>\n",
	 StartYear, StartMonth, StartDay, StartHour, StartMinute,
	 EndYear, EndMonth, EndDay, EndHour, EndMinute,
	 days
	 );
  fprintf(fp, "AccessLog contains (about) %.1f hits per day or (about) %.1f hits per hour for a total of %ld hits.\n<p>", 
	 LinesInLog/(float)days, LinesInLog/(float)(days*24), LinesInLog);

#ifdef DEBUG
  fprintf(stderr, "MYP2\n");
#endif

  if(ShowMaxFiles>0)
    fprintf(fp, "<a href=\"#files\">Files section</a><br>\n");

  if(ShowMaxReferers>0)
    fprintf(fp, "<a href=\"#referers\">Referers section</a><br>\n");

  if(ShowMaxBrowsers>0)
    fprintf(fp, "<a href=\"#browsers\">Browsers section</a><br>\n");
  
  if(ShowMaxOSnames>0)
    fprintf(fp, "<a href=\"#oses\">OS's section</a><br>\n");

  if(ShowMaxDomains>0)
    fprintf(fp, "<a href=\"#domains\">Domains section</a><br>\n");

  if(ShowMaxTopDomains>0)
    fprintf(fp, "<a href=\"#topdomains\">Topdomains section</a><br>\n");

  fprintf(fp, "<a href=\"#hourchart\">Hourchart section</a><br>\n");
/*  fprintf(fp, "<br>days = %d  LinesInLog = %d<br>\n", days, LinesInLog);*/

#ifdef DEBUG
  fprintf(stderr, "MYP3\n");
#endif

/* Removed due to privacy reasons... 
*/

/*   i=1; */
/*   while(i<20) */
/*     { */
/*       node=GetElement(&host, i); */
/*       fprintf(stderr, "node.name=%s hitcount=%ld\n", node->name, node->hitcount); */
/*       i++; */
/*     } */

  if(ShowMaxHosts>0)
    {
      fprintf(fp, "<br><a name=hosts><h4>Top Hosts (accesses) out of a total of %lu hosts</h4>\n", host.namecount);
      i=1;
      node=GetElement(&host, i);
      while( (node!=NULL) && (i<(ShowMaxHosts+1)))
	{
#ifdef DEBUG
  fprintf(stderr, "MYP3.1 namecount=%ld\n", host.namecount);
#endif
	  strcpy(TestName, "www.");
	  strcat(TestName, node->name);
#ifdef DEBUG
  fprintf(stderr, "MYP3.2\n");
#endif
	  if(HostExists(TestName))
	    fprintf(fp, "%d: <a href=\"http://%s\">%s</a> (%.2g%%/%lu)<br>\n", i, TestName, node->name, 100*node->hitcount/(float)LinesInLog, node->hitcount);
	  else
	    fprintf(fp, "%d: %s (%.2g%%/%lu)<br>\n", i, node->name, 100*node->hitcount/(float)LinesInLog, node->hitcount);
#ifdef DEBUG
  fprintf(stderr, "MYP3.3\n");
#endif
	  i++;
	  node=GetElement(&host, i);
	}
    }

#ifdef DEBUG
  fprintf(stderr, "MYP4\n");
#endif

  if(ShowMaxReferers>0)
    {
      fprintf(fp, "<br><a name=referers><h4>Top Referers (percentage) out of a total of %lu referers</h4>\n", referer.namecount);
      /* If most (=number one) references are "-" we do not include
	 it in the list otherwise we do */
      node=GetElement(&referer, 1);
      if(strncasecmp("-", node->name, 1)==0)
	{
	  i=2;
	  node=GetElement(&referer, i);
	  while((node!=NULL)&&(i<(ShowMaxReferers+2)))
	    {
	      fprintf(fp, "%d: <a href=\"%s\">", i-1, node->name);
	      htmlprint(fp, node->name, 60);
	      fprintf(fp, "</a> (%.2g%%/%lu)<br>\n", 100*node->hitcount/(float)LinesInLog, node->hitcount);
	      i++;
	      node=GetElement(&referer, i);
	    }
	}
      else
	{
	  i=1;
	  node=GetElement(&referer, i);
	  while((node!=NULL)&&(i<(ShowMaxReferers+1)))
	    {
	      fprintf(fp, "%d: ", i);
	      if(strncasecmp("-", node->name, 1)==0)
		fprintf(fp, "User went right here without nobody pointing...\n");
	      else
		htmlprint(fp, node->name, 60);
	      fprintf(fp, " (%.2g%%/%lu)<br>\n", 100*node->hitcount/(float)LinesInLog, node->hitcount);
	      i++;
	      node=GetElement(&referer, i);
	    }
	}
    }

#ifdef DEBUG
  fprintf(stderr, "MYP5\n");
#endif

  if(ShowMaxReferersites>0)
    {
      fprintf(fp, "<br><a name=referersites><h4>Top RefererSites (percentage) out of a total of %lu referersites</h4>\n", Referersites.namecount);
      /* If most (=number one) references are "-" we do not include
	 it in the list otherwise we do */
      node=GetElement(&Referersites, 1);
      if(strncasecmp("-", node->name, 1)==0)
	{
	  i=2;
	      node=GetElement(&Referersites, i);
	  while((node!=NULL)&&(i<(ShowMaxReferersites+2)))
	    {
	      fprintf(fp, "%d: <a href=\"%s\">", i-1, node->name);
	      htmlprint(fp, node->name, 60);
	      fprintf(fp, "</a> (%.2g%%/%lu)<br>\n", 100*node->hitcount/(float)LinesInLog, node->hitcount);
	      i++;
	      node=GetElement(&Referersites, i);
	    }
	}
      else
	{
	  i=1;
	  node=GetElement(&Referersites, i+1);
	  while((node!=NULL)&&(i<(ShowMaxReferersites+1)))
	    {
	      fprintf(fp, "%d: ", i);
	      if(strncasecmp("-", node->name, 1)==0)
		fprintf(fp, "User went right here without nobody pointing...\n");
	      else
		htmlprint(fp, node->name, 60);
	      fprintf(fp, " (%.2g%%/%lu)<br>\n", 100*node->hitcount/(float)LinesInLog, node->hitcount);
	      i++;
	      node=GetElement(&Referersites, i+1);
	    }
	}
    }

#ifdef DEBUG
  fprintf(stderr, "MYP6\n");
#endif
  
  if(ShowMaxBrowsers>0)
    {
      fprintf(fp, "<br><a name=browsers><h4>Top Browsers(uses) out of a total of %lu browsers</h4>\n", browser.namecount);
      i=1; 
      node=GetElement(&browser, i);
      while((node!=NULL)&&(i<(ShowMaxBrowsers+1))) 
	{
	  fprintf(fp, "%d: ", i); 
	  htmlprint(fp, node->name, 60); 
	  fprintf(fp, " (%.2g%%/%lu)<br>\n", 100*node->hitcount/(float)LinesInLog, node->hitcount); 
	  i++; 
	  node=GetElement(&browser, i);
	} 
    }
#ifdef DEBUG
  fprintf(stderr, "MYP7\n");
#endif

  if(ShowMaxOSnames>0)
    {
      fprintf(fp, "<br><a name=oses><h4>Top OS's (uses) out of a total of %lu OS's</h4>\n", OSname.namecount);
      i=1; 
      node=GetElement(&OSname, i);
      while((node!=NULL)&&(i<(ShowMaxOSnames+1))) 
	{
	  fprintf(fp, "%d: ", i); 
	  htmlprint(fp, node->name, 60); 
	  fprintf(fp, " (%.2g%%/%lu)<br>\n", 100*node->hitcount/(float)LinesInLog, node->hitcount); 
	  i++; 
	  node=GetElement(&OSname, i);
	} 
    }

#ifdef DEBUG
  fprintf(stderr, "MYP8\n");
#endif

  if(ShowMaxDomains>0)
    {
      fprintf(fp, "<br><a name=domains><h4>Top Domains (accesses) out of a total of %lu domains</h4>\n", domain.namecount);
      i=1;
      node=GetElement(&domain, i);
      while( (node!=NULL) && (i<(ShowMaxDomains+1)))
	{
	  strcpy(TestName, "www.");
	  strcat(TestName, node->name);
	  if(HostExists(TestName))
	    fprintf(fp, "%d: <a href=\"http://%s\">%s</a> (%.2g%%/%lu)<br>\n", i, TestName, node->name, 100*node->hitcount/(float)LinesInLog, node->hitcount);
	  else
	    fprintf(fp, "%d: %s (%.2g%%/%lu)<br>\n", i, node->name, 100*node->hitcount/(float)LinesInLog, node->hitcount);
	  i++;
	  node=GetElement(&domain, i);
	}
    }

#ifdef DEBUG
  fprintf(stderr, "MYP9\n");
#endif

  if(ShowMaxTopDomains>0)
    {
      fprintf(fp, "<br><a name=topdomains><h4>Top TopDomains (accesses) out of a total of %lu topdamains</h4>\n", TopDomains.namecount);
      i=1;
      node=GetElement(&TopDomains, i);
      while( (node!=NULL) && (i<(ShowMaxTopDomains+1)))
	{
	  strcpy(TestName, "www.");
	  strcat(TestName, node->name);
	  if(HostExists(TestName))
	    fprintf(fp, "%d: <a href=\"http://%s\">%s</a> (%.2g%%/%lu)<br>\n", i, TestName, node->name, 100*node->hitcount/(float)LinesInLog, node->hitcount);
	  else
	    fprintf(fp, "%d: %s (%.2g%%/%lu)<br>\n", i, node->name, 100*node->hitcount/(float)LinesInLog, node->hitcount);
	  i++;
	  node=GetElement(&TopDomains, i);
	}
    }

#ifdef DEBUG
  fprintf(stderr, "MYP10\n");
#endif

  /* Find Maximum of hits per hour*/
  i=0;
  max=0;
  while(i<24)
    {
      if(TimedAccess[i]>max)
	max=TimedAccess[i];
      i++;
    }

  value=max/50+1;
  fprintf(fp, "<br><a name=hourchart><h4>Accesses by the hour (local time of course)</h4><PRE>\n");
  i=0;
  while(i<24)
    {
      fprintf(fp, "%02d:00-%02d:59 ", i, i);
      j=TimedAccess[i];
      while((j>0)&&(j>value-1))
	{
	  fprintf(fp, "#");
	  j-=value;
	}
      fprintf(fp, " (%ld)\n", TimedAccess[i]);
      i++;
    }
  fprintf(fp, "            ---------+---------+---------+---------+---------+\n");
  fprintf(fp, "                    %3lu       %3lu       %3lu       %3lu       %3lu\n",
	 10*value,
	 20*value,
	 30*value,
	 40*value,
	 50*value);
  if(value>1)
    fprintf(fp, "</PRE><br>Each hash mark shows %lu accesses.<br>\n", value);
  else
    fprintf(fp, "</PRE><br>Each hash mark shows one access.<br>\n");

  /* Find Maximum of hits per month*/
  i=0;
  max=0;
  while(i<12)
    {
      if(MonthAccess[i]>max)
	max=MonthAccess[i];
      i++;
    }

  value=max/50+1;
  fprintf(fp, "<br><a name=monthchart><h4>Accesses by month</h4><PRE>\n");
  i=0;
  while(i<12)
    {
      fprintf(fp, "      ");
      MyMonthExpander(fp, i);
      j=MonthAccess[i];
      while((j>0)&&(j>value-1))
	{
	  fprintf(fp, "#");
	  j-=value;
	}
      fprintf(fp, " (%ld)\n", MonthAccess[i]);
      i++;
    }
  fprintf(fp, "            ---------+---------+---------+---------+---------+\n");
  fprintf(fp, "                    %3lu       %3lu       %3lu       %3lu       %3lu\n",
	 10*value,
	 20*value,
	 30*value,
	 40*value,
	 50*value);
  if(value>1)
    fprintf(fp, "</PRE><br>Each hash mark shows %lu accesses.<br>\n", value);
  else
    fprintf(fp, "</PRE><br>Each hash mark shows one access.<br>\n");

  /* Find Maximum of hits per day*/
  max=FindMaxCountList(&mdays);

  value=max/50+1;
  fprintf(fp, "<br><a name=dayschart><h4>Accesses by day</h4><PRE>\n");
  nodep=mdays.First;
  while(nodep!=NULL)
    {
      fprintf(fp, "      %s:", nodep->name);
      j=nodep->hitcount;
      while((j>0)&&(j>value-1))
	{
	  fprintf(fp, "#");
	  j-=value;
	}
      fprintf(fp, " (%ld)\n", nodep->hitcount);
      nodep=nodep->next;
    }
  fprintf(fp, "            ---------+---------+---------+---------+---------+\n");
  fprintf(fp, "                    %3lu       %3lu       %3lu       %3lu       %3lu\n",
	 10*value,
	 20*value,
	 30*value,
	 40*value,
	 50*value);
  if(value>1)
    fprintf(fp, "</PRE><br>Each hash mark shows %lu accesses.<br>\n", value);
  else
    fprintf(fp, "</PRE><br>Each hash mark shows one access.<br>\n");

/*   fprintf(fp, "<br><h4>Top Commands (file etc) (uses)</h4>\n"); */
/*   i=1; */
/*   while((i<=command.namecount)&&(i<(ShowMaxCommands+1))) */
/*     { */
/*       fprintf(fp, "%d: %s (%lu)<br>\n", i, command.name[i-1], command.hitcount[i-1]); */
/*       i++; */
/*     } */

#ifdef DEBUG
  fprintf(stderr, "MYP11\n");
#endif

  if(ShowMaxFiles>0)
    {
      fprintf(fp, "<br><a name=files><h4>Top requested \"files\" (requests) out of a total of %lu files</h4>\n", file.namecount);
      i=1;
      node=GetElement(&file, i);
      while((node!=NULL)&&(i<(ShowMaxFiles+1)))
	{
	  fprintf(fp, "%d: %s (%.2g%%/%lu)<br>\n", i, node->name, 100*node->hitcount/(float)LinesInLog, node->hitcount);
	  i++;
	  node=GetElement(&file, i);
	}
    }

  fprintf(fp, "<br><a name=summary><h4>Summary</h4>\n");
  fprintf(fp, "Number of hosts=%lu<br>\n", host.namecount);
  fprintf(fp, "Number of referers=%lu<br>\n", referer.namecount);
  fprintf(fp, "Number of referersites=%lu<br>\n", Referersites.namecount);
  fprintf(fp, "Number of browsers=%lu<br>\n", browser.namecount);
  fprintf(fp, "Number of domains=%lu<br>\n", domain.namecount);
  fprintf(fp, "Number of topdomains=%lu<br>\n", TopDomains.namecount);
  fprintf(fp, "Number of commands=%lu<br>\n", command.namecount);
  fprintf(fp, "Number of files=%lu<br>\n", file.namecount);
  fprintf(fp, "Number of OSnames=%lu<br>\n", OSname.namecount);
  fprintf(fp, "Number of days=%lu<br>\n", mdays.namecount);

#ifdef DEBUG
  fprintf(stderr, "MYP12\n");
#endif

  fprintf(fp, "<FORM method=GET action=\"http://altavista.digital.com/cgi-bin/query\">\n");
  fprintf(fp, "<INPUT TYPE=hidden NAME=pg VALUE=q>\n");
  fprintf(fp, "<INPUT TYPE=hidden NAME=what VALUE=web>\n");
  fprintf(fp, "<INPUT TYPE=hidden NAME=fmt VALUE=\"\">");
  fprintf(fp, "<INPUT TYPE=hidden NAME=q VALUE=\"+link:%s\">\n", PageName);
  fprintf(fp, "<INPUT TYPE=submit VALUE=\"Any links to this page ?\">\n");
  fprintf(fp, "</FORM>\n");

/* footnote */
  time(&tid);
  fprintf(fp, "<hr>This page was constructed %s", ctime(&tid));
  fprintf(fp, "<br>Created by <a href=\"http://www.lysator.liu.se/~jonass/programs.html\">%s</a> from ", PRG_NAME);
  fprintf(fp, "<a href=\"http://www.lysator.liu.se/~jonass/\">Jonas Svensson</a> <br>\n");
  fprintf(fp, "<a href=\"mailto:jonass@lysator.liu.se?subject=The Stat Page\">Comments ?</a><br>");
  fprintf(fp, "</body></html>\n");
/* end of footnote */
  if(fp!=stdout)
    fclose(fp);

/*   fprintf(stderr, "ShowMaxHosts=%d\n", ShowMaxHosts); */
/*   fprintf(stderr, "ShowMaxReferers=%d\n", ShowMaxReferers); */
/*   fprintf(stderr, "ShowMaxBrowsers=%d\n", ShowMaxBrowsers); */
/*   fprintf(stderr, "ShowMaxDomains=%d\n", ShowMaxDomains); */
/*   fprintf(stderr, "ShowMaxCommands=%d\n", ShowMaxCommands); */
/*   fprintf(stderr, "ShowMaxFiles=%d\n", ShowMaxFiles); */
/*   fprintf(stderr, "ShowMaxOSnames=%d\n", ShowMaxOSnames); */
/*   fprintf(stderr, "ShowMaxTopDomains=%d\n", ShowMaxTopDomains); */
/*   fprintf(stderr, "ShowMaxReferersites=%d\n", ShowMaxReferersites); */
}

/* find the name of the user/host */
char *FindHost(void)
{
  return buff;
}

void ip_cp(char *dest)
{
  char *src;

  src=FindHost();
  while((*src!='\0') && (*src!=' '))
    {
      *dest=*src;
      dest++;
      src++;
    }
}

char *TryNumToName(char *numaddr)
{
  struct hostent *hp;
  char local_buff[100];
  u_long addr;
 
  ip_cp(local_buff);

  if ((int)(addr = inet_addr(local_buff)) == -1) {
    /*fprintf(stderr, "FORMAT FAIL  %s", local_buff);*/
    return NULL;
  }
  
  hp = gethostbyaddr((char *)&addr, sizeof (addr), AF_INET);

  if(hp==NULL)
    {
      /*fprintf(stderr, "FAILED %s\n", local_buff);*/
      return NULL;
    }
  else
    {
      strcpy(HostNameTemp, hp->h_name);
/*      fprintf(stderr, "EUREKA %s -> %s\n", local_buff, HostNameTemp);*/
      return HostNameTemp;
    }
}

/* find out which second level domain user is from */
char *FindDomain(void)
{
  char *pek, *pek2;

  pek=FindHost();
  pek2=pek;
  if(isdigit(*pek))
    {
      pek2=TryNumToName(pek2);
      if(pek2!=NULL)
	pek=pek2;
    }
  /* Find end of string */
  while ((*pek!=' ')&&(*pek!='\0'))
    pek++;
  /* Find last dot */
  while (*pek!='.')
    pek--;
  pek--;
  /* Find next to last dot */
  while ((*pek!='.') && (*pek!='@'))
    pek--;
  /* discard dot */
  pek++;
  if(isdigit(*pek))
    return numerical;
  else
    return pek;
}

/* find out which topdomain visitor is from */
char *FindTopDomain(void)
{
  char *pek, *pek2;
  
  pek=FindHost();
  pek2=pek;
  if(isdigit(*pek))
    {
      pek2=TryNumToName(pek2);
      if(pek2!=NULL)
	pek=pek2;
    }
  /* Find end of string */
  while ((*pek!=' ')&&(*pek!='\0'))
    pek++;
  /* Find last dot */
  while (*pek!='.')
    pek--;
  pek++;
  
  if(isdigit(*pek))
    return numerical;
  else
    return pek;
}

/* which page made the user to request one of my pages */
char *FindReferer(void)
{
  char *t;

  t=FindHost();
  while(*t != ' ')
    t++;

  while(*t == ' ')
    t++;

  return t;
}

/* find the hostname of the referer */
char *FindRefererSite(void)
{
  char *t;

  t=FindReferer();
  if(*t=='h')
    {
      while(*t!='/')
	t++;
      t+=2;
    }
  return t;
}

/* find the name the browser (or what the browser said at least) */
char *FindBrowser(void)
{
  char *t;

  t=FindReferer();
  while(*t != ' ')
    t++;

  while(*t == ' ')
    t++;

  return t;
}

char *FindOSname(void)
{
  char *t;

  t=FindBrowser();
  while((*t != ' ')&&(*t != '('))
    t++;
  if(*t=='(')
    t++;
  if(*t==' ')
    t=noOS;
  else
    {
      if(strncasecmp(t, "compatible", strlen("compatible"))==0)
	{
	  /* Skip "compatible" */
	  while((*t!=';')&&(*t!=' ')&&(*t!='\0'))
	    {
	      t++;
	    }
	  if(*t==';')
	    t++;
	  /* Skip "MSIE" */
	  while((*t!=';')&&(*t!=' ')&&(*t!='\0'))
	    {
	      t++;
	    }
	  if(*t==';')
	    t++;
	}
    }
  return t;
}

char *FindDate(void)
{
  char *t;
  int state=0;
  
  t=FindHost();
  while(state<10)
    {
      t++;
      if(t=='\0')
	state=12;
      switch(state)
	{
	case 0 :
	  if(*t==' ') state=1;
	  break;
	case 1 :
	  if(*t == '[')
	    state=11;
	  else
	    state=0;
	  break;
	}
    }

  if(state==11)
    t++; /* drop the [ */
  else
    t=NULL;
  return t;
}

char *FindCommand(void)
{
  char *t;

  t=FindDate();
  while(*t != ']')
    t++;

  t++; /* drop the ] */

  while(*t == ' ')
    t++;

  t++; /* drop the quote-char */

  return t;
}

char *FindFileName(void)
{
  char *t;

  t=FindCommand();
/*  fprintf(stderr, "\n0: %s", t);*/
  while(*t != ' ')
    t++;
/*  fprintf(stderr, "1: %s", t);*/

  while(*t == ' ')
    t++;
/*   fprintf(stderr, "2: %s", t); */

  while(*t != ' ')
    t++;
/*   fprintf(stderr, "3: %s", t); */
  t--;
  t--; /* if the filename ends in a slash */
  while(*t != '/')
    t--;

  t++;
/*   fprintf(stderr, "4: %s", t); */

  return t;
}


unsigned long GetYear(void)
{
  unsigned long t;

  sscanf(FindDate()+7, "%lu", &t);
  return t;
}

unsigned long GetMonth(void)
/* returns 1 for january, 2 for february etc. */
{
  unsigned long t=15;

  if(strncasecmp("jan", FindDate()+3, 3)==0)
    t=1;
  if(strncasecmp("feb", FindDate()+3, 3)==0)
    t=2;
  if(strncasecmp("mar", FindDate()+3, 3)==0)
    t=3;
  if(strncasecmp("apr", FindDate()+3, 3)==0)
    t=4;
  if(strncasecmp("may", FindDate()+3, 3)==0)
    t=5;
  if(strncasecmp("jun", FindDate()+3, 3)==0)
    t=6;
  if(strncasecmp("jul", FindDate()+3, 3)==0)
    t=7;
  if(strncasecmp("aug", FindDate()+3, 3)==0)
    t=8;
  if(strncasecmp("sep", FindDate()+3, 3)==0)
    t=9;
  if(strncasecmp("oct", FindDate()+3, 3)==0)
    t=10;
  if(strncasecmp("nov", FindDate()+3, 3)==0)
    t=11;
  if(strncasecmp("dec", FindDate()+3, 3)==0)
    t=12;


  if(t==15)
    printf("<<<<<<<<<<<< E R R O R GetMonth %c%c%c>>>>>>>>>>>>>>>\n"
           ,*(FindDate()+3)
           ,*(FindDate()+4)
           ,*(FindDate()+5));
  return t;
}

unsigned long GetDay(void)
{
  unsigned long t;

  sscanf(FindDate(), "%lu", &t);
  return t;
}

unsigned long GetHour(void)
{
  unsigned long t;

  sscanf(FindDate()+12, "%lu", &t);
  return t;
}

unsigned long GetMinute(void)
{
  unsigned long t;

  sscanf(FindDate()+15, "%lu", &t);
  return t;
}

int CompareDate(unsigned long Year1,
		unsigned long Month1,
		unsigned long Day1,
		unsigned long Hour1,
		unsigned long Minute1,
		unsigned long Year2,
		unsigned long Month2,
		unsigned long Day2,
		unsigned long Hour2,
		unsigned long Minute2)
/* returns  1 if the first date is after the second 
           -1                      before
*/
{
  if(Year1>Year2)
    return 1;
  else
    if(Year1<Year2)
      return -1;
    else
      if(Month1>Month2)
	return 1;
      else
	if(Month1<Month2)
	  return -1;
	else
	  if(Day1>Day2)
	    return 1;
	  else
	    if(Day1<Day2)
	      return -1;
	    else
	      if(Hour1>Hour2)
		return 1;
	      else
		if(Hour1<Hour2)
		  return -1;
		else
		  if(Minute1>Minute2)
		    return 1;
		  else
		    if(Minute1<Minute2)
		      return -1;
		    else
		      return 0;
}

void SetNewStartDate(void)
{
  StartYear=GetYear();
  StartMonth=GetMonth();
  StartDay=GetDay();
  StartHour=GetHour();
  StartMinute=GetMinute();
}

void SetNewEndDate(void)
{
  EndYear=GetYear();
  EndMonth=GetMonth();
  EndDay=GetDay();
  EndHour=GetHour();
  EndMinute=GetMinute();
}

void CopySimpleName(char *src, char *dest, char endchar)
{
  if(*src==' ')
    strcpy(dest, "Unknown (tm)");
  else
    {
      while((*src!=endchar)&&(*src!='\0')&&(*src!=' '))
	{
	  *dest=*src;
	  dest++;
	  src++;
	}
      *dest='\0';
    }
}

void CopyReferersiteName(char *src, char *dest)
{
  if(*src=='-')
    strcpy(dest, "None (tm)");
  else
    {
      while((*src!='/')&&(*src!='\0')&&(*src!=' ')&&(*src!=':'))
	{
	  *dest=*src;
	  dest++;
	  src++;
	}
      *dest='\0';
    }
}

void CopyBrowserName(char *src, char *dest)
{
  char *realdest;
  realdest=dest;

#ifdef DEBUG
  fprintf(stderr, "first=%s", src);
#endif

  while((*src!='(')&&(*src!=' ')&&(*src!='\0'))
    {
      *dest=*src;
      dest++;
      src++;
    }
#ifdef DEBUG
  fprintf(stderr, "h=%s", src);
#endif
  src++;

  if(strncasecmp(src, "compatible", strlen("compatible"))==0)
    {
#ifdef DEBUG
      fprintf(stderr, "hej\n");
#endif
      *dest=' ';
      dest++;
      /* Copy "compatible" */
      while((*src!=';')&&(*src!=' ')&&(*src!='\0'))
	{
	  *dest=*src;
	  dest++;
	  src++;
	}
      if(*src==';')
	src++;
      /* Copy "MSIE" */
      while((*src!=';')&&(*src!=' ')&&(*src!='\0'))
	{
	  *dest=*src;
	  dest++;
	  src++;
	}
      if(*src==';')
	src++;
    }
  *dest='\0';
#ifdef DEBUG
  fprintf(stderr, "realdest=%s\n", realdest);
#endif
}

void DoTheStuff(void)
{
  unsigned long t;
  /*  char *charpek; */

#ifdef DEBUG
      fprintf(stderr, "DTS1\n");
#endif
  
  LinesInLog++;
  if(
     CompareDate(GetYear(), GetMonth(), GetDay(), GetHour(), GetMinute(),
		 StartYear, StartMonth, StartDay, StartHour, StartMinute)
     ==-1)
    SetNewStartDate();
  
  if(
     CompareDate(GetYear(), GetMonth(), GetDay(), GetHour(), GetMinute(),
		 EndYear, EndMonth, EndDay, EndHour, EndMinute)
     ==1)
    SetNewEndDate();
  
  
#ifdef DEBUG
      fprintf(stderr, "DTS2\n");
#endif
  TimedAccess[GetHour()]++;
  MonthAccess[GetMonth()-1]++;

#ifdef DEBUG
      fprintf(stderr, "DTS3\n");
#endif
  
  CopySimpleName(FindHost(), temp, ' '); 
  t=NameInList(temp, &host); 
  if(t==0)  
    {  
      AddNameToList(temp, &host);
    }  
  else 
    IncCount(&host, t);

  CopySimpleName(FindDate(), temp, ':'); 
  t=NameInList(temp, &mdays);
  if(t==0)
    {
      AddNameToList(temp, &mdays);
    }
  else
    IncCount(&mdays, t);

#ifdef DEBUG
      fprintf(stderr, "DTS4 %s\n", buff);
#endif
  
  CopySimpleName(FindReferer(), temp, ' ');
#ifdef DEBUG
      fprintf(stderr, "DTS4b\n");
#endif
      /* hr dumpar den kor */
  t=NameInList(temp, &referer); 
  /* hit kommer den inte */
#ifdef DEBUG
      fprintf(stderr, "DTS4c\n");
#endif
  if(t==0) 
    { 
#ifdef DEBUG
      fprintf(stderr, "DTS4d\n");
#endif
      AddNameToList(temp, &referer); 
#ifdef DEBUG
      fprintf(stderr, "DTS4e\n");
#endif
    } 
  else 
    { 
#ifdef DEBUG
      fprintf(stderr, "DTS4f\n");
#endif
    IncCount(&referer, t); 
#ifdef DEBUG
      fprintf(stderr, "DTS4g\n");
#endif
    } 

#ifdef DEBUG
      fprintf(stderr, "DTS5\n");
#endif
  
  CopyReferersiteName(FindRefererSite(), temp);
  t=NameInList(temp, &Referersites); 
  if(t==0) 
    { 
      AddNameToList(temp, &Referersites); 
    } 
  else 
    IncCount(&Referersites, t); 

#ifdef DEBUG
      fprintf(stderr, "DTS6\n");
#endif
  
  /*    fprintf(stderr, "\nFB=%s", FindBrowser()); */
  CopyBrowserName(FindBrowser(), temp); 
#ifdef DEBUG
  fprintf(stderr, "CBN1\n");
#endif
  t=NameInList(temp, &browser);
#ifdef DEBUG
  fprintf(stderr, "CBN2\n");
#endif
  if(t==0) 
    { 
#ifdef DEBUG
      fprintf(stderr, "CBN3\n");
#endif
      AddNameToList(temp, &browser); 
#ifdef DEBUG
      fprintf(stderr, "CBN4\n");
#endif
    } 
  else 
    { 
#ifdef DEBUG
      fprintf(stderr, "CBN5\n");
#endif
      IncCount(&browser, t); 
#ifdef DEBUG
      fprintf(stderr, "CBN6\n");
#endif
    } 
  
#ifdef DEBUG
  fprintf(stderr, "COS1\n");
#endif
  CopySimpleName(FindOSname(), temp, ')'); 
#ifdef DEBUG
  fprintf(stderr, "COS2\n");
#endif
  t=NameInList(temp, &OSname);
#ifdef DEBUG
  fprintf(stderr, "COS3\n");
#endif
  if(t==0) 
    { 
#ifdef DEBUG
      fprintf(stderr, "COS4\n");
#endif
      AddNameToList(temp, &OSname); 
#ifdef DEBUG
      fprintf(stderr, "COS5\n");
#endif
    } 
  else 
    { 
#ifdef DEBUG
      fprintf(stderr, "COS6\n");
#endif
      IncCount(&OSname, t); 
#ifdef DEBUG
      fprintf(stderr, "COS7\n");
#endif
    } 
  
#ifdef DEBUG
      fprintf(stderr, "CD1\n");
#endif
  CopySimpleName(FindDomain(), temp, ' '); 
  t=NameInList(temp, &domain); 
  if(t==0) 
    { 
      AddNameToList(temp, &domain); 
    } 
  else 
    IncCount(&domain, t); 

#ifdef DEBUG
      fprintf(stderr, "CD2\n");
#endif

#ifdef DEBUG
      fprintf(stderr, "CTD1\n");
#endif
  CopySimpleName(FindTopDomain(), temp, ' '); 
  t=NameInList(temp, &TopDomains); 
  if(t==0) 
    { 
      AddNameToList(temp, &TopDomains); 
    } 
  else 
    IncCount(&TopDomains, t); 

#ifdef DEBUG
      fprintf(stderr, "CTD26\n");
#endif

#ifdef DEBUG
      fprintf(stderr, "CC1\n");
#endif
  CopySimpleName(FindCommand(), temp, '\"'); 
  t=NameInList(temp, &command); 
  if(t==0) 
    { 
      AddNameToList(temp, &command); 
    } 
  else 
    IncCount(&command, t); 
#ifdef DEBUG
      fprintf(stderr, "CC2\n");
#endif

#ifdef DEBUG
      fprintf(stderr, "CFN1\n");
#endif
  CopySimpleName(FindFileName(), temp, ' '); 
  t=NameInList(temp, &file); 
  if(t==0) 
    { 
      AddNameToList(temp, &file); 
    } 
  else 
    IncCount(&file, t); 
#ifdef DEBUG
      fprintf(stderr, "CFN2\n");
#endif
/*   fprintf(stderr, "line: %sFindFileName:%s", buff, FindFileName()); */
}

int main()
{
  FILE *fp;
  
  init();
  
  fp=fopen(LoggFileName, "r");
  if(fp==NULL)
    {
      printf("could not open \"%s\"\n", LoggFileName);
      exit(1);
    }
  fgets(buff, MAX_LINE_LENGTH-1, fp);
  while(!feof(fp))
    {
#ifdef DEBUG
      fprintf(stderr, "main1\n");
#endif
      if(strlen(buff)>20 )
	{
#ifdef DEBUG
      fprintf(stderr, "main2\n");
#endif
	  DoTheStuff();
#ifdef DEBUG
      fprintf(stderr, "main3\n");
#endif
	}
#ifdef DEBUG
      fprintf(stderr, "main4\n");
#endif
      fgets(buff, MAX_LINE_LENGTH-1, fp);
#ifdef DEBUG
      fprintf(stderr, "main5\n");
#endif
    }
  fclose(fp);

#ifdef DEBUG
      fprintf(stderr, "main6\n");
#endif
  SortNames(&host);
  SortNames(&referer);
  SortNames(&browser);
  SortNames(&domain);
  SortNames(&command);
  SortNames(&file);
  SortNames(&OSname);
  SortNames(&TopDomains);
  SortNames(&Referersites);
#ifdef DEBUG
      fprintf(stderr, "main7\n");
#endif

  MyPrintSummary();
  clean();
  return 0;
}








