/* see inimodule.h for instructions or http://www.lysator.liu.se/~jonass */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "inimodule.h"

void ExpandSectionName(char *src, char *dest)
{
  sprintf(dest, "[%s]", src);
}

BOOL ReadFindSection(FILE *fp, char *Section, char *tempbuffer)
{
  BOOL found;
  char ExpandedSectionName[INI_MAX_LINE_LENGTH];

  found=FALSE;

  ExpandSectionName(Section, ExpandedSectionName);

#ifdef DEBUG
  fprintf(stderr, "ReadFindSection1: ->%s<-\n", ExpandedSectionName);
#endif

  fgets(tempbuffer, INI_MAX_LINE_LENGTH, fp);
  while( (!found) && (!feof(fp)))
    {
#ifdef DEBUG
      fprintf(stderr, "ReadFindSection examines: %s\n", tempbuffer);
#endif
      if(tempbuffer[0]=='[') /* snabbare n strncmp */
	{
	  if(strncmp(tempbuffer, ExpandedSectionName, strlen(ExpandedSectionName))==0)
	    {
	      found=TRUE;
	    }
	  else
	    {
	      fgets(tempbuffer, INI_MAX_LINE_LENGTH, fp);
	    }
	}
      else
	{
	  fgets(tempbuffer, INI_MAX_LINE_LENGTH, fp);
	}
    }
#ifdef DEBUG
  if(found==TRUE)
    fprintf(stderr, "ReadFindSection returns TRUE\n");
  else
    fprintf(stderr, "ReadFindSection returns FALSE\n");
#endif

  return found;
}

BOOL ReadFindKey(FILE *fp, char *KeyName, char *tempbuffer)
{
  BOOL found, newsection;
  char ExpandedKeyName[INI_MAX_LINE_LENGTH];

  found=FALSE;
  newsection=FALSE;

  strncpy(ExpandedKeyName,
	  KeyName,
	  INI_MAX_LINE_LENGTH-1
	 );

  strcat(ExpandedKeyName, "=");
#ifdef DEBUG
  fprintf(stderr, "ReadFindKey1: ->%s<-\n", ExpandedKeyName);
#endif

  fgets(tempbuffer, INI_MAX_LINE_LENGTH, fp);
  while( (!found) && (!feof(fp)) && (!newsection))
    {
      if(tempbuffer[0]!='[') /* snabbare n strncmp */
	{
	  if(strncmp(tempbuffer, ExpandedKeyName, strlen(ExpandedKeyName))==0)
	    {
	      found=TRUE;
	    }
	  else
	    {
	      fgets(tempbuffer, INI_MAX_LINE_LENGTH, fp);
	    }
	}
      else
	{
	  newsection=TRUE;
	}
    }
  return found;
}

int ExtractIntValue(char *tempbuffer, int DefaultValue)
{
  /* skip to equalsign or end of line */
  while( (*tempbuffer!='=') && (*tempbuffer!='\n'))
    tempbuffer++;

  /* skip equal sign */
  if(*tempbuffer=='=')
    tempbuffer++;

  /* skip leading spaces */
  while(*tempbuffer==' ')
    tempbuffer++;

  /* now we have found either end of line or the value */
  if(*tempbuffer=='\n')
    return DefaultValue;
  else
    return atoi(tempbuffer);
}

int GetPrivateProfileInt(char *SectionName, char *KeyName, int DefaultValue, char *FileName)
{
  FILE *fp;
  int returnvalue;
  char tempbuffer[INI_MAX_LINE_LENGTH];
 
  fp=fopen(FileName, "r");
  if(fp==NULL)
    {
      returnvalue=DefaultValue;
    }
  else
    {
      if(FALSE==ReadFindSection(fp, SectionName, tempbuffer))
	{
	  returnvalue=DefaultValue;
	}
      else
	{
	  if(FALSE==ReadFindKey(fp, KeyName, tempbuffer))
	    {
	      returnvalue=DefaultValue;
	    }
	  else
	    {
	      returnvalue=ExtractIntValue(tempbuffer, DefaultValue);
	    }
	}
      fclose(fp);
    }
  return returnvalue;
}

int ExtractStringValue(char *tempbuffer, char *dest)
{
  char *srcpek, *destpek;

  /* skip to equalsign or end of line */
  while( (*tempbuffer!='=') && (*tempbuffer!='\n'))
    tempbuffer++;

  /* skip equal sign */
  if(*tempbuffer=='=')
    tempbuffer++;

  /* skip leading spaces */
  while(*tempbuffer==' ')
    tempbuffer++;

  /* now we have found either end of line or the value */
  srcpek=tempbuffer;
  destpek=dest;
  while( (*srcpek!='\0') && (*srcpek!='\n'))
    {
      *destpek=*srcpek;
      srcpek++;
      destpek++;
    }
  return strlen(dest);
}

int GetPrivateProfileString(char *SectionName, char *KeyName, char *DefaultValue, char *Buffer, int BufferSize, char *FileName)
{
  FILE *fp;
  BOOL result;
  int returnvalue;
  char tempbuffer[INI_MAX_LINE_LENGTH];
 
  result=FALSE;
  
  fp=fopen(FileName, "r");
  if(fp!=NULL)
    {
      if(TRUE==ReadFindSection(fp, SectionName, tempbuffer))
	{
	  if(TRUE==ReadFindKey(fp, KeyName, tempbuffer))
	    {
	      returnvalue=ExtractStringValue(tempbuffer, Buffer);
	      result=TRUE;
	    }
	}
      fclose(fp);
    }
  if(result==FALSE)
    strcpy(Buffer, DefaultValue);
  return strlen(Buffer);
}

int GetProfileInt(char *SectionName, char *KeyName, int DefaultValue)
{
  return GetPrivateProfileInt(SectionName, KeyName, DefaultValue, USERS_SYSTEM_FILENAME);
}

int GetProfileString(char *SectionName, char *KeyName, char *DefaultValue, char *Buffer, int BufferSize, char *FileName)
{
  return GetPrivateProfileString(SectionName, KeyName, DefaultValue, Buffer, BufferSize, USERS_SYSTEM_FILENAME);
}



BOOL WriteFindSection(FILE *fpsrc, FILE *fpdest, char *Section, char *tempbuffer)
{
  BOOL found;
  char ExpandedSectionName[INI_MAX_LINE_LENGTH];

  found=FALSE;
  
  ExpandSectionName(Section, ExpandedSectionName);

#ifdef DEBUG
  fprintf(stderr, "WriteFindSection1: ->%s<-\n", ExpandedSectionName);
#endif
  
  fgets(tempbuffer, INI_MAX_LINE_LENGTH, fpsrc);
  while( (!found) && (!feof(fpsrc)))
    {
      fputs(tempbuffer, fpdest);
      if(tempbuffer[0]=='[') /* snabbare n strncmp */
	{
	  if(strncmp(tempbuffer, ExpandedSectionName, strlen(ExpandedSectionName))==0)
	    {
	      found=TRUE;
	    }
	  else
	    {
	      fgets(tempbuffer, INI_MAX_LINE_LENGTH, fpsrc);
	    }
	}
      else
	{
	  fgets(tempbuffer, INI_MAX_LINE_LENGTH, fpsrc);
	}
    }
  return found;
}


BOOL WritePrivateProfileInt(char *SectionName, char *KeyName, int Value, char *FileName)
{
  char localbuff[INI_MAX_LINE_LENGTH];

  sprintf(localbuff, "%d", Value);

  return WritePrivateProfileString(SectionName, KeyName, localbuff, FileName);
}

BOOL WriteFindKey(FILE *fpsrc, FILE *fpdest, char *KeyName, char *tempbuffer)
/* reads to end of section or up to and including key */
{
  BOOL found, newsection;
  char ExpandedKeyName[INI_MAX_LINE_LENGTH];

  found=FALSE;
  newsection=FALSE;

  strncpy(ExpandedKeyName,
	  KeyName,
	  INI_MAX_LINE_LENGTH-1
	 );

  strcat(ExpandedKeyName, "=");
#ifdef DEBUG
  fprintf(stderr, "WriteFindKey1: ->%s<-\n", ExpandedKeyName);
#endif
  
  fgets(tempbuffer, INI_MAX_LINE_LENGTH, fpsrc);
  while( (!found) && (!feof(fpsrc)) && (!newsection))
    {
      if(tempbuffer[0]!='[') /* snabbare n strncmp */
	{
	  if(strncmp(tempbuffer, ExpandedKeyName, strlen(ExpandedKeyName))==0)
	    {
	      found=TRUE;
	    }
	  else
	    {
	      fputs(tempbuffer, fpdest);
	      fgets(tempbuffer, INI_MAX_LINE_LENGTH, fpsrc);
	    }
	}
      else
	{
	  newsection=TRUE;
	}
    }
  return found;
}

BOOL CopyRestOfFile(FILE *src, FILE *dest)
{
  char buff[INI_MAX_LINE_LENGTH];
  
  if(!feof(src))
    fgets(buff, INI_MAX_LINE_LENGTH, src);
  while(!feof(src))
    {
      fputs(buff, dest);
      fgets(buff, INI_MAX_LINE_LENGTH, src);
    }
  return TRUE; /* always succed ...  */
}

BOOL WritePrivateProfileString(char *SectionName, char *KeyName, char *Value, char *FileName)
{
  FILE *fpsrc, *fptemp;
  BOOL result;
  char tempbuffer[INI_MAX_LINE_LENGTH];

  result=FALSE;
  
  fptemp=tmpfile();
  if(fptemp!=NULL)
    {
      fpsrc=fopen(FileName, "r");
      if(fpsrc!=NULL)
	{
	  if(TRUE==WriteFindSection(fpsrc, fptemp, SectionName, tempbuffer))
	    {
	      if(TRUE==WriteFindKey(fpsrc, fptemp, KeyName, tempbuffer))
		{
		  fprintf(fptemp, "%s=%s\n", KeyName, Value);
		  result=TRUE;
		}
	      else
		{
		  fprintf(fptemp, "%s=%s\n", KeyName, Value);
		  fputs(tempbuffer, fptemp);
		  result=TRUE;
		}
	    }
	  else /* did not find section */
	    {
	      fprintf(fptemp, "[%s]\n", SectionName);
	      fprintf(fptemp, "%s=%s\n", KeyName, Value);
	      result=TRUE;
	    }
	  CopyRestOfFile(fpsrc, fptemp);
	  fclose(fpsrc);
	  rewind(fptemp);
	  fpsrc=fopen(FileName, "w");
	  if(fpsrc==NULL)
	    {
	      fprintf(stderr, "Could not open %s for writing (WritePrivateProfileString:1)\n", FileName);
	      result=FALSE;
	    }
	  else
	    {
	      CopyRestOfFile(fptemp, fpsrc);
	      fclose(fpsrc);
	    }
	}
      else
	{
	  fpsrc=fopen(FileName, "w");
	  if(fpsrc==NULL)
	    {
	      fprintf(stderr, "Could not open %s for writing(WritePrivateProfileString:2)\n", FileName);
	      result=FALSE;
	    }
	  else
	    {
	      fprintf(fpsrc, "[%s]\n", SectionName);
	      fprintf(fpsrc, "%s=%s\n", KeyName, Value);
	      fclose(fpsrc);
	      result=TRUE;
	    }
	}
      fclose(fptemp);
    }
  else
    {
      fprintf(stderr, "Could not open temporary file(WritePrivateProfileString:3)\n");
    }

  return result;
}

BOOL WriteProfileInt(char *SectionName, char *KeyName, int Value)
{
  return WritePrivateProfileInt(SectionName, KeyName, Value, USERS_SYSTEM_FILENAME);
}

BOOL WriteProfileString(char *SectionName, char *KeyName, char *Value)
{
  return WritePrivateProfileString(SectionName, KeyName, Value, USERS_SYSTEM_FILENAME);
}


